# GNU MathProg version of the investment problem,
# using a node-based formulation of a scenario tree
#
# This file includes the model and output


# The scenario tree
param LastNode;                      # index of the last node
set NODES := 0 .. LastNode;          # all nodes in the scenario tree
param Root in NODES default 0;       # root of the tree
set FUT_NODES := NODES diff {Root};  # all nodes except the root
param ChildPerNode default 2;        # number of children nodes of each node
param Pred{n in FUT_NODES} default (n-1) div ChildPerNode; # predecessors
param FirstLeaf;                     # index of the 1st leaf (last-stage node)
set LEAVES := FirstLeaf .. LastNode; # set of leaves
param Prob{LEAVES} default 1/(LastNode - FirstLeaf + 1); # probabilities

# Deterministic model and parameters
set ASSETS;       # set of assets
param InitBudget; # initial budget
param CapTarget;  # capital target

# Stochastic parameters
param Return{FUT_NODES, ASSETS};

# Stochastic model variables
var x{NODES diff LEAVES, ASSETS} >= 0; # investment
var w{LEAVES} >= 0;                    # shortage
var y{LEAVES} >= 0;                    # surplus

# Objective function
minimize shortage_minus_surplus:
	sum{n in LEAVES} Prob[n] * (1.3*w[n] - 1.1*y[n]);


## CONSTRAINTS
subject to initial_budget:
	sum{a in ASSETS} x[Root, a] = InitBudget;

subject to cashflow_balance{n in FUT_NODES diff LEAVES}:
	sum{a in ASSETS} Return[n, a] * x[Pred[n], a] = sum{a in ASSETS} x[n, a];

subject to final_balanace{n in LEAVES}:
	sum{a in ASSETS} Return[n, a] * x[Pred[n], a] + w[n] - y[n] = CapTarget;


## EXTRA SETS USED IN THE OUTPUT
set PRED{n in FUT_NODES}           # set of all predecessors of each node
	:= {Pred[n]} union (if Pred[n] = Root then {} else PRED[Pred[n]]);
set SC_NODES{s in 1..card(LEAVES)} # set of nodes in each scenario
	:= setof{n in NODES diff LEAVES: n in PRED[s + FirstLeaf - 1]} n
	   union {s + FirstLeaf - 1}; 


## -----------------------------------------------------------
#data investment.dat # uncomment to solve the model with AMPL
#display PRED > nul; # needed by AMPL!

solve;

# Formatted output
printf "\nOptimal objective value = %g\n\n",
	sum{n in LEAVES} Prob[n] * (1.3*w[n] - 1.1*y[n]);

printf "x :=";
printf{a in ASSETS}: "\t%s ", a;
printf "\n";
for{n in NODES diff LEAVES} {
	printf "%5d ", n;
	printf{a in ASSETS}: "\t%6.2f ", x[n,a];
	printf "\n";
}

printf "\nWealth as a function of time (target is %g):\n", CapTarget;
for{s in 1..card(LEAVES)} {
	printf "sc. %d: ", s;
	printf{n in SC_NODES[s] diff LEAVES} "%.1f -> ", sum{a in ASSETS} x[n, a];
	printf{n in SC_NODES[s] inter LEAVES} "%5.1f\n", CapTarget - w[n] + y[n];
}
printf "\n";

end;
