// Implementation of the investment model, using FlopC++
// This version uses a simple deteriministic equivalent

#include <iomanip>
#include "flopc.hpp"

// Change these two lines to use a different solver
#include "OsiClpSolverInterface.hpp"
#define OSI_SOLVER_INTERFACE OsiClpSolverInterface

using namespace std;
using namespace flopc;


// Predecessor of a given node; input and output as a number
int get_pred(const int node) {
	return floor((node-1) / 2);
}
// Predecessor of a given node; input and output as an MP_index
// This means that this can be used directly in index expressions!
MP_index_exp get_pred(MP_index &node) {
	return floor((node-1) / 2);
}

int main()
{
	// DATA - This would be normally read from some external file
	enum {stocks, bonds, nmbAssets}; // assets to invest to; sets nmbAssets = 2
	int Root = 0;      // root node
	int nmbNodes = 15; // nodes are 0..14
	int firstLeaf = 7; // leaves are 7..14

	double InitBudget = 55;
	double CapTarget = 80;

	double retData[][nmbAssets] = {
		{1.25, 1.14},
		{1.06, 1.16},
		{1.21, 1.17},
		{1.07, 1.12},
		{1.15, 1.18},
		{1.06, 1.12},
		{1.26, 1.13},
		{1.07, 1.14},
		{1.25, 1.15},
		{1.06, 1.12},
		{1.05, 1.17},
		{1.06, 1.15},
		{1.05, 1.14},
		{1.06, 1.12}
	};

	// initialize the object for the model
	MP_model &model = MP_model::getDefaultModel();
	model.setSolver(new OSI_SOLVER_INTERFACE);
	#ifdef NDEBUG
		model.silent();  // non-debug, i.e. release version -> less output
	#else
		model.verbose(); // debug version -> show more output
	#endif

	// Sets
	MP_set NODES(nmbNodes);
	MP_subset<2> PREDS(NODES,NODES);  // A `set of predecessors' for each node
	MP_subset<1> FUT_NODES(NODES);    // All nodes except Root
	MP_subset<1> LEAVES(NODES);       // Leaves of the tree
	MP_subset<1> BRANCH_NODES(NODES); // All nodes except leaves
	MP_subset<1> MID_NODES(NODES);    // All nodes except Root and leaves
	MP_set ASSETS(nmbAssets);
	MP_index n, a, pr; // for use with other sets, see bellow

	// Fill the subsets; shows both the direct syntax and using an MP_index
	forall (NODES.such_that(NODES != Root), FUT_NODES.insert(NODES));
	forall (NODES(n).such_that(n >= firstLeaf), LEAVES.insert(n));
	forall (NODES(n).such_that(n < firstLeaf), BRANCH_NODES.insert(n));
	forall (NODES(n).such_that(n != Root && n < firstLeaf), MID_NODES.insert(n));
	assert (FUT_NODES.size() == nmbNodes - 1 && "Check size of set FUT_NODES");
	assert (LEAVES.size() == nmbNodes - firstLeaf && "Check size of set LEAVES");
	assert (BRANCH_NODES.size() == firstLeaf && "Check size of set BRANCH_NODES");
	assert (MID_NODES.size() == firstLeaf - 1 && "Check size of set MID_NODES");
	//PREDS.display("PREDS");
	//FUT_NODES.display("FUT_NODES");
	//LEAVES.display("LEAVES");
	//BRANCH_NODES.display("BRANCH_NODES");
	//MID_NODES.display("MID_NODES");

	// Data - tree structure
	/* For general tree, the predecessors would be given in data. In our case,
	   however, we have a regular binary tree, so we can use a formula.
	   Here, we use the version returning MP_index_exp, so we can use it
	   directly in the 'forall' expression. Alternatively, we could have used
	   the version returning a number and write
	   : for (int i = 1; i < nmbNodes; i++) PREDS.insert(i, get_pred(i)); */
	forall (NODES(n).such_that(n != Root), PREDS.insert(n, get_pred(n)));
	//PREDS.display("PREDS");

	// Probabilities of the leaves
	MP_data Prob(LEAVES);
	Prob(LEAVES) = 1.0 / LEAVES.size();

	// Print the node info. Note how to get an index of a node in the subset
	cout << endl << "Node information:" << endl;
	for (int i = 0; i < NODES.size(); i++) {
		cout << "k = " << setw(2) << i
		     << "  index(FUT_N) = " << setw(2) << FUT_NODES(i);
		if (FUT_NODES(i) >= 0) {
			for (int j = 0; j < firstLeaf; j++)
				// check if 'j' is a predecessor of 'i'
				if (PREDS(i,j) >= 0)
					cout << "  Pred(k) = " << j;
		} else {
			cout << "  Pred(k) = " << "-";
		}
		cout << "  index(LEAVES) = " << setw(2) << LEAVES(i);
		if (LEAVES(i) >= 0)
			cout << "  Prob(k) = " << Prob(LEAVES(i));
		else
			cout << "  Prob(k) = " << " -";
		cout << endl;
	}
	cout << endl;

	// Init data object and read data; alt.: MP_data Return(&retData[0][0]);
	MP_data Return(FUT_NODES, ASSETS);
	Return.value(&retData[0][0]); // use table retData, startin from el.[0][0]

	// Variables
	/* We could set the lower bound to zero using "w.lowerLimit(LEAVES) = 0;"
	   but it is not necessary, as zero is the default value
	   (Note how we have made the definition even more AMPL-like by splitting
	    the line and thus avoiding repeating MP_variable for each variable.) */
	MP_variable
		x(BRANCH_NODES, ASSETS), // investment
		w(LEAVES),               // shortage
		y(LEAVES);               // surplus

	// Objective - define it as an expression; later say it should be minimized
	MP_expression objFunc
		= sum(LEAVES, Prob(LEAVES) * (1.3 * w(LEAVES) - 1.1 * y(LEAVES)));

	// Constraints
	MP_constraint
		initial_budget,
		cashflow_balance(MID_NODES),
		final_balance(LEAVES);

	initial_budget = sum(ASSETS(a), x(Root, a)) == InitBudget;

	/* This is more tricky: we have to realize that the different variables and
	   parameters used are defined on different sets, to we have to calculate
	   index in the set the variable/parameter is actually defined on.
	   This constraint shows how to use the PREDS set: we simply sum over all the
	   predecessors of node 'n' - so we better be sure that there is only one.
	   Note that while it would seem more logical to store the predecessors in
	   MP_data Pred(NODES), this is not possible since a variable can not be
	   indexed by an MP_data object. */
	cashflow_balance(MID_NODES(n))
		= sum(ASSETS(a)*PREDS(n,pr), Return(FUT_NODES(n), a)
		                             * x(BRANCH_NODES(pr), a))
		  == sum(ASSETS(a), x(BRANCH_NODES(n), a));

	/* This version illustrates direct use of the get_pred function. For this,
	   we need the version returning MP_index_exp. */
	final_balance(LEAVES(n))
		= sum(ASSETS(a), Return(FUT_NODES(n), a) * x(BRANCH_NODES(get_pred(n)), a))
		  + w(LEAVES(n)) - y(LEAVES(n)) == CapTarget;

	// Set the objective; attach the model;
	model.setObjective(objFunc);
	model.attach();

	// Write MPS file
	model->writeMps("investment");

	// Minimize the objective, ie. solve the model; alt: model.minimize(objFunc);
	model.minimize();

	return 0;
}
