// Implementation of the investment model, using COIN SMI

#include "SmiScnModel.hpp"

// Change these two lines to use a different solver
#include "OsiClpSolverInterface.hpp"
#define OSI_SOLVER_INTERFACE OsiClpSolverInterface

using namespace std;


// Forward declaration of functions (implementations follows after 'main()')

/// Create OSI object with a deterministic version of the model
void create_det_model(OsiSolverInterface &modelOSI);

/// Add a scenario to the SMI stochastic model
int add_scenario(SmiScnModel *ptStochModel, SmiCoreData *ptCoreModel,
                 double *matrixElem, int brStage, int parScen, double prob);


int main()
{
	// First, we build the CORE problem, i.e. the deterministic version
	OSI_SOLVER_INTERFACE detModel;
	create_det_model(detModel); // Create the deterministic model

	// Next, add information about stages and create the SMI core data object
	int nmbStages = 4;                          // Number of stages
	int colStages[] = {0, 0, 1, 1, 2, 2, 3, 3}; // Stages of columns/variables
	int rowStages[] = {0, 1, 2, 3};             // Stages of rows/constraints

	// alt.: SmiCoreData *ptStochCore = new SmiCoreData(&detModel, ...);
	SmiCoreData stochCore(&detModel, nmbStages, colStages, rowStages);


	/* Finally, add the scenario tree.
	   This is done in an SMPS-like fashion, ie. each scenario has a parent
	   scenario it branches from. We then have to specify the branching stage
	   and all the data that are different than the parent's.
	   In our case, the only difference is in the matrix A. We only need to
	   specify the elements that differ from the parent, that is the returns.
	*/
	// Vector of the matrix elements that differ from the prev. scenario
	// (Note that we waste some space here, declaring all vectors to be [6].)
	double scenDataRet[][6] = {
		{ -1.25, -1.14, -1.21, -1.17, -1.26, -1.13 }, // scenario ending at node 7
		{                             -1.07, -1.14 }, // scenario ending at node 8
		{               -1.07, -1.12, -1.25, -1.15 }, // scenario ending at node 9
		{                             -1.06, -1.12 }, // scenario ending at node 10
		{ -1.06, -1.16, -1.15, -1.18, -1.05, -1.17 }, // scenario ending at node 11
		{                             -1.06, -1.15 }, // scenario ending at node 12
		{               -1.06, -1.12, -1.05, -1.14 }, // scenario ending at node 13
		{                             -1.06, -1.12 }  // scenario ending at node 14
	};
	int branchStage[] = { 1, 3, 2, 3, 1, 3, 2, 3 }; // Branching stage
	int nmbScen = 8;                                // Number of scenarios
 	double scenProb = 1.0 / nmbScen;                // Scenarios' probability

	// Create the SMI stochastic model object
	// alt.: SmiScnModel *ptStochModel = new SmiScnModel();
	SmiScnModel stochModel;

	for (SmiScenarioIndex sc = 0; sc < nmbScen; sc++) {
		int parentScen = (sc == 0 ? 0 : sc - 1); // Parent scenario
		#ifdef NDEBUG
			add_scenario(&stochModel, &stochCore, scenDataRet[sc], branchStage[sc],
		               parentScen, scenProb);
		#else
			int newScenIndx = add_scenario(&stochModel, &stochCore, scenDataRet[sc],
																		 branchStage[sc], parentScen, scenProb);
			assert (newScenIndx == sc && "Check index of the new scenario");
		#endif
	}


	// ------------------------------------------------------------------------ //
	// Now, the stochastic model is complete.
	/* Unfortunately, there is no stochastic solver in COIN-OR yet, so we have to
	   solve the model using a deterministic solver on the determ. equivalent. */

	// Attach a solver to the stochastic model
	stochModel.setOsiSolverHandle(*new OSI_SOLVER_INTERFACE());

	// 'loadOsiSolverData()' loads the deterministic equivalent into an
	// internal OSI data structure and returns a handle (pointer) to it.
	OsiSolverInterface *ptDetEqModel = stochModel.loadOsiSolverData();

	// 'ptDetEqModel' now includes the deterministic equivalent
	ptDetEqModel->writeMps("investment.det-equiv");
	cout << endl << "Solving the deterministic equivalent:" << endl;
	ptDetEqModel->initialSolve();
	cout << endl << "The deterministic equivalent model has "
	     << ptDetEqModel->getNumCols() << " variables and "
	     << ptDetEqModel->getNumRows() << " constraints." << endl;
	cout << "Optimal objective value = " << ptDetEqModel->getObjValue() << endl;


	// ------------------------------------------------------------------------ //
	/* Even if we use a deterministic solver, we can still get information about
	   the solution on the scenario tree, using the SMI (stochastic) model */
	cout << endl << "The stochastic model has " << stochModel.getNumScenarios()
	     << " scenarios." << endl;
	assert (stochModel.getNumScenarios() == nmbScen && "Check number of scens.");

	// We will calculate the wealth at each node of the tree, plus the obj. value
	vector<double> nodeWealth(nmbStages, 0);
	double objValue = 0.0;

	// Get vectors of variable values and their objective coefficients
	const double *ptOsiSolution = ptDetEqModel->getColSolution();

	// Compute the wealth at each node, by traversing the tree from leafs up
	for (SmiScenarioIndex sc = 0; sc < nmbScen; sc ++) {
		// Get the leaf node of scenario sc:
		SmiScnNode *ptNode = stochModel.getLeafNode(sc);
		int nodeStage = nmbStages;

		double scProb = ptNode->getModelProb(); // probability of the leaf
		double scenObjVal = stochModel.getObjectiveValue(sc); // objective value
		objValue += scProb * scenObjVal;
		printf ("scen %d: prob = %.3f  obj =%7.2f", sc+1, scProb, scenObjVal);

		// This loop traverses the tree, from the leaf to the root
		while (ptNode != NULL) {
			// info about columns of ptNode in the OSI model (ptDetEqModel)
			int startColInOsi = ptNode->getColStart();
			int nmbColsInOsi = ptNode->getNumCols();

			// Loop through the variables corresponding to this node
			double wealth = 0.0;
			if (nodeStage < nmbStages) {
				// The node is not a leaf -> just sum up the variables
				for (int j = startColInOsi; j < startColInOsi + nmbColsInOsi; j++) {
					wealth += ptOsiSolution[j];
				}
			} else {
				// Special rules for the leaves
				assert (nmbColsInOsi == 2 && "Leaves have only two variables (w,y)");
				assert (ptNode->getNumRows() == 1 && "Leaves have only one row");

				// get RHS of the only constraint for this node (get LB, check LB=UB)
				double capTarget = -ptDetEqModel->getRowLower()[ptNode->getRowStart()];
				assert (-capTarget == ptDetEqModel->getRowUpper()[ptNode->getRowStart()]
				        && "Leaves have an equality constraint, so LB=UB");

				// In the leaves, the wealth is computed as: -w + y + capTarget
				wealth = -ptOsiSolution[startColInOsi]
				         + ptOsiSolution[startColInOsi+1]
				         + capTarget;
			}
			nodeWealth[nodeStage-1] = wealth;

			// Get the parent node (Root will return NULL, stopping the loop)
			ptNode = ptNode->getParent();
			nodeStage--;
		}

		printf (";  wealth:");
		for (int t = 0; t<nmbStages-1; t++)
			printf ("%6.2f ->", nodeWealth[t]);
		printf ("%6.2f\n", nodeWealth[nmbStages-1]);
	}

	printf ("%15s Total obj = %7.3f\n", "", objValue);
	assert (fabs(objValue - ptDetEqModel->getObjValue()) < 1e-6 && "Check obj.");

	return 0;
}


// -------------------------------------------------------------------------- //
// -------------------------------------------------------------------------- //

// Create OSI object with a deterministic version of the model
// The return values do not matter, so we use the expected values
void create_det_model(OsiSolverInterface &modelOSI)
{
	// Matrix A, using the CoinPackedMatrix (sparse matrix) format
	/* Non-zero elements given by three vectors: row and column indices + values
	   (not the most efficient way, but easy to read) */
	int rowIndx[] = { // Row indices of the elements
		0, 0,
		1, 1, 1, 1,
		      2, 2, 2, 2,
		            3, 3, 3, 3
	};
	int colIndx[] = { // Column indices of the elements
		0, 1,
		0, 1, 2, 3,
		      2, 3, 4, 5,
		            4, 5, 6, 7
	};
	double elem[] = { // Values of the non-zero elements (expected value problem)
		 1,      1,
		-1.155, -1.150,  1,      1,
				        	  -1.123, -1.148,  1,      1,
					                          -1.108, -1.140, -1,  1
	};
	CoinBigIndex nmbElem = 14; // Number of non-zero elements
	bool storeByCols = false;  // Internal parameter, both values should work
	CoinPackedMatrix A(storeByCols, rowIndx, colIndx, elem, nmbElem);

	// Bounds for variables are not needed, as the default is [0,inf]
	// Bounds on the rows, i.e. RHS (equality constraints, so LB=UB)
	double rowLB[] = {55, 0, 0, -80};
	double rowUB[] = {55, 0, 0, -80};

	// Objective function
	double objCoef[] = {0, 0, 0, 0, 0, 0, 1.3, -1.1};

	// Load the problem to the provided OSI handle
	// (Note that we do not really have to create A, we could use directly
	// CoinPackedMatrix (storeByCols, rowIndx, colIndx, elem, nmbElem) as an arg.)
	modelOSI.loadProblem(A, NULL, NULL, objCoef, rowLB, rowUB);

	// Optional: set names and write an MPS file
	const char *colNames[] = {"x11", "x12", "x21", "x22", "x31", "x32", "w", "y"};
	const char *rowNames[] = {"st1", "st2", "st3", "st4"};
	modelOSI.writeMpsNative("investment.core.mps", rowNames, colNames);
	cout << endl << "The core (deterministic) model has " << modelOSI.getNumCols()
	     << " variables and " << modelOSI.getNumRows() << " constraints." << endl;
}


// -------------------------------------------------------------------------- //

// Add a scenario to the SMI stochastic model
int add_scenario(SmiScnModel *ptStochModel, SmiCoreData *ptCoreModel,
                 double *matrixElem, int brStage, int parScen, double prob)
{
	/* The differing matrix elements are again described by the triplets
	   (row index, column index, value), given in three vectors.
	   The indices depend on the branching stage, but in such a way that indices
	   for branching at a later stage are subsets of indices for branching at
	   stage 2. As a result, we only need one vector...  */
	// Row and column indices, for branching at stage 2
	int rowDifIndx[] = {1, 1, 2, 2, 3, 3};
	int colDifIndx[] = {0, 1, 2, 3, 4, 5};
	int indxStart = 2 * (brStage-1); // Where to start reading the index vectors
	int nmbElem = 6 - indxStart;     // Number of elements in the matrix

	// Now create the matrix containing the elements that differ from the parent
	/* Note that the default action is to replace the old values by the new ones,
	   but other actions may be used as well (addition, multiplication, etc.) */
	CoinPackedMatrix ADiff(false,
	                       &(rowDifIndx[indxStart]), &(colDifIndx[indxStart]),
	                       matrixElem, nmbElem);

	// Add the scenario to the SMI object.
	/* Since the objective and the bounds on rows and columns do not change at
	   all, we just send empty pointers. */
	int scenIndx = ptStochModel->generateScenario(ptCoreModel, &ADiff,
	                                              NULL, NULL, NULL, NULL, NULL,
	                                              brStage, parScen, prob);

	// Optional message
	cout << "Added scenario " << scenIndx << ", total number of scenarios is "
	     << ptStochModel->getNumScenarios() << endl;

	return scenIndx; // Return the index of the new scenario
}
