// Implementation of the investment model, using COIN SMI

#include <iomanip>
#include <sstream>
#include <deque>
#include "SmiScnModel.hpp"
#include "SmiScnData.hpp"

// Change these two lines to use a different solver
#include "OsiClpSolverInterface.hpp"
#define OSI_SOLVER_INTERFACE OsiClpSolverInterface

using namespace std;


int main(int argc, char* argv[])
{
	if (argc != 2) {
		cerr << endl
		     << "ERROR: This program needs one argument, the base name of the smps."
		     << endl
		     << "       The input is then expected to be in files:" << endl
		     << "       <name>.core, <name>.time, and <name>.stoch," << endl
		     << "       where the extensions may be truncated to 4 or 3 characters."
		     << endl;
		return 1;
	}

	SmiScnModel stochModel;

	// read SMPS model from files <name>.core, <name>.time, and <name>.stoch
	int fileReadOK = stochModel.readSmps(argv[1]);
	if (fileReadOK < 0) {
		cerr << endl << "ERROR: "
				 << "There was some problem reading the SMPS file ... aborting" << endl;
		return fileReadOK;
	}

	// Attach a solver to the stochastic model
	//OSI_SOLVER_INTERFACE *ptOsiSolverObj = new OSI_SOLVER_INTERFACE();
	//stochModel.setOsiSolverHandle(*ptOsiSolverObj);
	stochModel.setOsiSolverHandle(* new OSI_SOLVER_INTERFACE());

	// 'loadOsiSolverData()' loads the deterministic equivalent into an internal
	// OSI data structure and returns a handle to it.
	OsiSolverInterface *ptDetEqModel = stochModel.loadOsiSolverData();
	cout << endl << "The deterministic equivalent model has "
	     << ptDetEqModel->getNumCols() << " variables and "
	     << ptDetEqModel->getNumRows() << " constraints." << endl;

	// Write MPS of the the deterministic equivalent
	stringstream fileNameSS; // for easy file name creation; alt. use sprintf
	fileNameSS << argv[1] << ".det-equiv";
	ptDetEqModel->writeMps(fileNameSS.str().c_str());


	// -------------------------------------------------------------------------
	// Optional: set some nice Hints to the OSI solver
	ptDetEqModel->setHintParam(OsiDoPresolveInInitial,true);
	ptDetEqModel->setHintParam(OsiDoScale,true);

	// Solve the deterministic equivalent
	cout << endl << "Solving the deterministic equivalent:" << endl;
	ptDetEqModel->initialSolve();

	// Basic output
	cout << endl << "Optimal objective value = " << ptDetEqModel->getObjValue()
			 << endl;


	// -------------------------------------------------------------------------
	/* Even if we use a deterministic solver, we still get information about the
	   solution on the scenario tree, using the SMI (stochastic) model */
	int nmbScen = stochModel.getNumScenarios();
	cout << endl << "The stochastic model has " << nmbScen << " scenarios."
	     << endl;

	// Vector of solution to the deterministic equivalent
	const double *ptOsiSolution = ptDetEqModel->getColSolution();

	// Compute sum of variables at each node of the tree
	// (just to show how to access the variable values)
	deque<double> sumNodeVars;

	// Compute the objective value as the mean of scenario values
	double objValue = 0.0;

	// Traversing the tree by scenarios, reporting some values
	for (SmiScenarioIndex sc = 0; sc < nmbScen; sc ++) {
		// Get the leaf node of scenario sc:
		SmiScnNode *ptNode = stochModel.getLeafNode(sc);

		double scProb = ptNode->getModelProb(); // probability of the leaf
		double scenObjVal = stochModel.getObjectiveValue(sc); // objective value
		objValue += scProb * scenObjVal;
		printf ("scen %d: prob = %.3f  obj =%7.2f", sc+1, scProb, scenObjVal);

		// This loop traverses the tree, from the leaf to the root
		while (ptNode != NULL) {
			// info about columns of ptNode in the OSI model (ptDetEqModel)
			int startColInOsi = ptNode->getColStart();
			int nmbColsInOsi = ptNode->getNumCols();

			// Loop through the variables corresponding to this node, compute a sum
			// (Does not make much sense, just to show how to access variables.)
			double sum = 0.0;
			for (int j = startColInOsi; j < startColInOsi + nmbColsInOsi; j++) {
				sum += ptOsiSolution[j];
			}
			sumNodeVars.push_front(sum);

			// Get the parent node (Root will return NULL, stopping the loop)
			ptNode = ptNode->getParent();
		}
		int nmbStages = sumNodeVars.size();

		printf (";  sum(var):");
		for (int t = 0; t<nmbStages-1; t++)
			printf ("%6.2f ->", sumNodeVars[t]);
		printf ("%6.2f\n", sumNodeVars[nmbStages - 1]);

		sumNodeVars.clear();
	}

	printf ("%15s Total obj = %7.3f\n", "", objValue);
	assert (fabs(objValue - ptDetEqModel->getObjValue()) < 1e-6
	        && "Check of the manually computed objective value");

	return 0;
}
